# Copyright (C) 2021 Hyunwoong Ko <kevin.ko@tunib.ai> and Sang-Kil Park <skpark1224@hyundai.com>
# All rights reserved.

import logging
from logging import getLogger

logging.basicConfig(format="[Kss]: %(message)s", level=logging.WARNING)
logger = getLogger()
